
def sortLines(script, pageStart):
    newScript = []
    for page in script:
        if page["page"] < pageStart:
            continue
        newScript.append({
            "page": page["page"],
            "content": []
        })

        newScript[-1]["content"] = page["content"]

        newScript[-1]["content"].sort(
            key=lambda curr: (curr["y"], curr["x"]))

        # TODO: how to determine this?
        for i, content in enumerate(newScript[-1]["content"]):
            if abs(content["y"] - newScript[-1]["content"][i-1]["y"]) < 5:
                newScript[-1]["content"][i-1]["y"] = content["y"]
        # print(content)
        # print(newScript[-1]["content"][i-1])

        newScript[-1]["content"].sort(
            key=lambda curr: (curr["y"], curr["x"]))

    return newScript



import json
import re

from screenplay_pdf_to_json.utils import isCharacter, isParenthetical, extractCharacter, isHeading, extractHeading

LAST_SCENE = -2


def groupSections(topTrends, script, pageStart):
    """group types into the same sections"""

    newScript = categorizeSections(topTrends, script, pageStart)
    newScript = combineCategories(newScript, pageStart)
    newScript = divideParentheticals(newScript)
    return newScript


def divideParentheticals(newScript):
    """seperates parentheticals from dialogues"""

    for page in newScript:
        for i, section in enumerate(page["content"]):
            for j, scene in enumerate(section["scene"]):
                if scene["type"] == "CHARACTER":
                    scene["content"]["dialogue"] = getParenthetical(
                        scene["content"]["dialogue"])
                elif scene["type"] == "DUAL_DIALOGUE":
                    scene["content"]["character1"]["dialogue"] = getParenthetical(
                        scene["content"]["character1"]["dialogue"])
                    scene["content"]["character2"]["dialogue"] = getParenthetical(
                        scene["content"]["character2"]["dialogue"])
    return newScript


def getParenthetical(text):
    """splits dialogue string into list, seperating any containing parenthetical(s)"""

    return list(
        filter(lambda x: len(x.strip()) > 0, re.split(r'(\([^)]+\))', text))
    )


def combineCategories(newScript, pageStart):
    """combines consecutive sections with the same type"""

    finalSections = []
    for page in newScript:
        if page["page"] < pageStart:
            continue
        finalSections.append({"page": page["page"], "content": []})

        for i, content in enumerate(page["content"]):
            finalSections[-1]["content"].append({
                "scene_number": content["scene_number"],
                "scene_info": content["scene_info"],
                "scene": []
            })
            j = 0
            while j < len(content["scene"]):
                scene = content["scene"][j]

                sectionSameTypeAsPrevious = j > 0 and scene["type"] == content["scene"][j-1]["type"]
                if scene["type"] == "CHARACTER":
                    finalSections[-1]["content"][-1]["scene"].append({
                        "type": "CHARACTER",
                        "content": {
                            "character": scene["text"]["character"],
                            "modifier": scene["text"]["modifier"],
                            "dialogue": ""
                        }
                    })
                elif scene["type"] == "DUAL_DIALOGUE":
                    sectionIsCharacter = lambda currScene, character: len(currScene["content"][character]) == 1 and isCharacter(currScene["content"][character][0])
                    if sectionIsCharacter(scene, "character1") and sectionIsCharacter(scene, "character2"):
                        finalSections[-1]["content"][-1]["scene"].append({
                            "type": "DUAL_DIALOGUE",
                            "content": {
                                "character1": extractCharacter(scene["content"]["character1"][0]),
                                "character2":  extractCharacter(scene["content"]["character2"][0]),
                            }
                        })
                    else:
                        finalSections[-1]["content"][-1]["scene"][-1]["content"]["character1"]["dialogue"] = getJoinedText(
                            scene["content"]["character1"])
                        finalSections[-1]["content"][-1]["scene"][-1]["content"]["character2"]["dialogue"] = getJoinedText(
                            scene["content"]["character2"])

                elif scene["type"] == "DIALOGUE" and content["scene"][j-1]["type"] == "CHARACTER":
                    finalSections[-1]["content"][-1]["scene"][-1]["content"]["dialogue"] += scene["text"]
                elif sectionSameTypeAsPrevious and scene["type"] == "DIALOGUE":
                    finalSections[-1]["content"][-1]["scene"][-1]["content"]["dialogue"] += " " + scene["text"]
                elif sectionSameTypeAsPrevious and scene["type"] == "ACTION":
                    # if part of same paragraph, concat text
                    if (scene["content"][0]["y"] - finalSections[-1]["content"][-1]["scene"][-1]["content"][-1]["y"] <= 16):
                        finalSections[-1]["content"][-1]["scene"][-1]["content"][-1]["text"] += " " + scene["content"][0]["text"]
                        finalSections[-1]["content"][-1]["scene"][-1]["content"][-1]["y"] = scene["content"][0]["y"]
                    # else, just append entire text
                    else:
                        finalSections[-1]["content"][-1]["scene"][-1]["content"].append(
                            scene["content"][0])
                else:
                    finalSections[-1]["content"][-1]["scene"].append(scene)

                j += 1
    return finalSections


def getJoinedText(textArr):
    return " ".join([arr["text"]
                     for arr in textArr])


def categorizeSections(topTrends, script, pageStart):
    """categorize lines into types"""

    finalSections = []
    sceneNumber = 0
    for page in script:
        if page["page"] < pageStart:
            continue
        finalSections.append({"page": page["page"], "content": []})

        finalSections[-1]["content"].append({
            "scene_number": sceneNumber,
            "scene_info": finalSections[LAST_SCENE]["content"][-1]["scene_info"] if len(finalSections) >= 2 else None,
            "scene": []
        })

        characterOccurred = False
        for i, content in enumerate(page["content"]):
            if "character2" in content:
                finalSections[-1]["content"][-1]["scene"].append({
                    "type": "DUAL_DIALOGUE",
                    "content": {
                        "character1": content["segment"],
                        "character2": content["character2"],
                    }
                })
                characterOccurred = False
                continue

            previousY = page["content"][i -
                                        1]["segment"][-1]["y"] if i > 0 else 0
            x = content["segment"][0]["x"]
            y = content["segment"][0]["y"]
            text = content["segment"][0]["text"]

            # booleans
            isTransition = content["segment"][0]["x"] >= 420 or "FADE IN:" in text
            isAction = abs(x - topTrends[0][0]) <= 15

            if isHeading(content["segment"][0]):
                sceneNumber += 1
                if len(finalSections[-1]["content"][-1]["scene"]) == 0:
                    finalSections[-1]["content"][-1] = {
                        "scene_number": sceneNumber,
                        "scene_info": extractHeading(content["segment"][0]["text"]),
                        "scene": []
                    }
                else:
                    finalSections[-1]["content"].append({
                        "scene_number": sceneNumber,
                        "scene_info": extractHeading(content["segment"][0]["text"]),
                        "scene": []
                    })
                characterOccurred = False
            elif isTransition:
                finalSections[-1]["content"][-1]["scene"].append({
                    "type": "TRANSITION",
                    "content": {
                        "text": text,
                        "metadata": {
                            "x": x,
                            "y": y
                        }
                    }
                })
                characterOccurred = False
            elif isAction:
                # if Heading is multi-line
                if i > 0 and len(finalSections[-1]["content"][-1]["scene"]) == 0 and y - page["content"][i-1]["segment"][-1]["y"] < 24:
                    finalSections[-1]["content"][-1]["scene_info"]["location"] += " " + text
                else:
                    finalSections[-1]["content"][-1]["scene"].append({
                        "type": "ACTION",
                        "content": [{"text": text, "x": x, "y": y}]
                    })
                    characterOccurred = False

            elif isCharacter(content["segment"][0]):
                finalSections[-1]["content"][-1]["scene"].append({
                    "type": "CHARACTER",
                    "text": extractCharacter(content["segment"][0]),
                    "metadata": {
                        "x": x,
                        "y": y
                    }
                })
                characterOccurred = True
            else:
                currentScene = finalSections[-1]["content"][-1]["scene"]

                # first line of page is never a dialogue
                if len(currentScene) == 0 or not characterOccurred:
                    finalSections[-1]["content"][-1]["scene"].append({
                        "type": "ACTION",
                        "content": [{"text": text, "x": x, "y": y}]
                    })

                else:
                    finalSections[-1]["content"][-1]["scene"].append({
                        "type": "DIALOGUE",
                        "text": text,
                        "metadata": {
                            "x": x,
                            "y": y
                        }
                    })

    return finalSections



from .groupLines import parsePdf
from .groupDualDialogues import groupDualDialogues
from .groupSections import groupSections
from .SortLines import sortLines
from .cleanPage import cleanPage
from .getTopTrends import getTopTrends
from .stitchSeperateWordsIntoLines import stitchSeperateWordsIntoLines
from .processInitialPages import processInitialPages



from screenplay_pdf_to_json.utils import isCharacter
import json
import copy

LATEST_PAGE = -1
EPSILON = 3


def groupDualDialogues(script, pageStart):
    """detects and groups dual dialogues"""

    newScript = []
    for page in script:
        if page["page"] < pageStart:
            continue
        newScript.append({"page": page["page"], "content": []})

        i = 0
        isDualDialogue = 0

        while i < len(page["content"]):
            content = page["content"][i]
            currentY = round(content["y"])
            segmentToAdd = [{
                "x": content["x"],
                "y": content["y"],
                "text": content["text"]
            }]

            nextContent = page["content"][i +
                                          1] if i + 1 < len(page["content"]) else False

            prevContent = page["content"][i - 1] if i -1 >= 0  else False

            previousIsCharacter = prevContent and isCharacter(prevContent)

            # if current line and next line is character, and previous line is not character (3 characters in a row is impossible)
            if not previousIsCharacter and  nextContent and isCharacter(content) and isCharacter(nextContent):
                isDualDialogue = 1

            # if next content is the same line, and isDualDialogue > 0, then it's a dual dialogue
            if nextContent and currentY == nextContent["y"] and isDualDialogue > 0:
                character2ToAdd = {
                    "x": nextContent["x"],
                    "y": nextContent["y"],
                    "text": nextContent["text"]
                }
                left = segmentToAdd[0]
                right = character2ToAdd
                if left["x"] > nextContent["x"]:
                    right = left
                    left = copy.copy(right)

                if isDualDialogue <= 2:
                    newScript[-1]["content"].append({
                        "segment": [left],
                        "character2": [right]
                    })
                else:
                    newScript[-1]["content"][-1]["segment"].append(left)
                    newScript[-1]["content"][-1]["character2"].append(right)
                i += 1
                isDualDialogue += 1

            # if content resides in a different y axis, we know it's not part of a dual dialogue
            else:
                isDualDialogue = 0
                # add content's y axis as key and the content array index position as value
                newScript[-1]["content"].append({
                    "segment": segmentToAdd
                })
            i += 1

    newScript = stitchLastDialogue(newScript, pageStart)
    return newScript


def stitchLastDialogue(script, pageStart):
    """
    detect last line of a dual dialogue. This isn't detected by detectDualDialogue since
    a dialogue may be longer than the other, and therefore take up a different y value
    """
    currScript = []
    for page in script:
        if page["page"] < pageStart:
            continue
        currScript.append({"page": page["page"], "content": []})
        margin = -1
        for i, content in enumerate(page["content"]):
            # if margin > 0, then content is potentially a dual dialogue
            if margin > 0:
                currScriptLen = len(currScript[LATEST_PAGE]["content"]) - 1

                # content might be the last line of dual dialogue, or not
                if "character2" not in content and i > 0:
                    # last line of a dual dialogue
                    if abs(content["segment"][0]["y"] - page["content"][i-1]["segment"][LATEST_PAGE]["y"]) <= margin + EPSILON:
                        def getDiff(contentX, currX): return abs(
                            contentX - currX)

                        diffBetweenContentAndSegment = getDiff(
                            content["segment"][0]["x"], currScript[LATEST_PAGE]["content"][currScriptLen]["segment"][0]["x"])
                        diffBetweenContentAndCharacter2 = getDiff(
                            content["segment"][0]["x"], currScript[LATEST_PAGE]["content"][currScriptLen]["character2"][0]["x"]) if "character2" in currScript[LATEST_PAGE]["content"][currScriptLen] else -1

                        if diffBetweenContentAndSegment < diffBetweenContentAndCharacter2:
                            currScript[LATEST_PAGE]["content"][currScriptLen]["segment"] += content["segment"]
                        else:
                            currScript[LATEST_PAGE]["content"][currScriptLen]["character2"] += content["segment"]

                    # not a dual dialogue. fuk outta here!
                    else:
                        currScript[LATEST_PAGE]['content'].append(content)
                        margin = 0

                # still a dual dialogue
                else:
                    currScript[LATEST_PAGE]["content"].append(content)

            # if no dual
            else:
                if "character2" in content:
                    # margin between character head and FIRST line of dialogue
                    margin = abs(page["content"][i+1]["segment"][0]["y"] -
                                 content["segment"][LATEST_PAGE]["y"])
                    currScript[LATEST_PAGE]['content'].append(content)
                else:
                    currScript[LATEST_PAGE]['content'].append(content)

    return currScript



from sys import stderr

def isParenthetical(text):
    return "(" in text[0] and ")" in text[-1]


def extractCharacter(currentContent):
    text = currentContent["text"]
    split = text.split()
    modifier = text[text.find(
        "(")+1:text.find(")")] if text.find("(") != -1 else None
    character = text.replace(
        "("+modifier+")", "") if modifier is not None else text
    return {
        "character": character,
        "modifier": modifier,
    }


def isCharacter(currentContent):
    text = currentContent["text"]
    characterNameEnum = ["V.O", "O.S", "CONT'D"]

    if isParenthetical(text):
        return False

    for heading in characterNameEnum:
        if heading in text:
            return True

    if not text[0].isalpha() and "\"" not in text[0]:
        return False

    if text != text.upper():
        return False

    # check if header?
    if any(x in text for x in ["--", "!", "?", "@", "%"]):
        return False

    if any(x in text[-1] for x in ["-", "."]):
        return False


    return True



from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.pdfpage import PDFPage
from pdfminer.converter import XMLConverter, HTMLConverter, TextConverter
from pdfminer.layout import LAParams, LTTextBox
from pdfminer.pdfdocument import PDFDocument
from pdfminer.pdfparser import PDFParser
from pdfminer.converter import PDFPageAggregator
from pdfminer.pdfdevice import PDFDevice

import io
import pdfminer
import unicodedata

import json


def parsePdf(scriptFile):
    newScript = {
        "pdf": []
    }
    # Create a PDF parser object associated with the file object.
    parser = PDFParser(scriptFile)

    # Create a PDF document object that stores the document structure.
    document = PDFDocument(parser)

    # Check if the document allows text extraction. If not, abort.
    if not document.is_extractable:
        return print("wrong format")

    # Create a PDF resource manager object that stores shared resources.
    rsrcmgr = PDFResourceManager()

    # Create a PDF device object.
    device = PDFDevice(rsrcmgr)

    # BEGIN LAYOUT ANALYSIS
    # Set parameters for analysis.
    laparams = LAParams(word_margin=100, boxes_flow=1,
                        line_margin=0.3)

    # Create a PDF page aggregator object.
    device = PDFPageAggregator(rsrcmgr, laparams=laparams)

    # Create a PDF interpreter object.
    interpreter = PDFPageInterpreter(rsrcmgr, device)

    i = 0
    # loop over all pages in the document
    for page in PDFPage.create_pages(document):
        newScript["pdf"].append({
            "page": i,
            "content": []
        })

        # read the page into a layout object
        interpreter.process_page(page)
        layout = device.get_result()

        # extract text from this object
        parseObj(newScript, layout._objs, page.mediabox[3])
        i += 1
    return newScript
    

def parseObj(newScript, lt_objs, pageHeight):

    # loop over the object list
    for obj in lt_objs:

        if isinstance(obj, pdfminer.layout.LTTextLine):
            # text = obj.get_text()
            # if text.strip():
            #     for c in obj._objs:
            #         if isinstance(c, pdfminer.layout.LTChar):
            #             print("fontname %s" % c)

            newScript["pdf"][-1]["content"].append({
                "x": round(obj.bbox[0]),
                "y": round(pageHeight - obj.bbox[1]),
                "text": obj.get_text().replace('\n', '').strip()
            })
        # if it's a textbox, also recurse
        if isinstance(obj, pdfminer.layout.LTTextBoxHorizontal):
            parseObj(newScript, obj._objs, pageHeight)



import json
import re
from screenplay_pdf_to_json.parse_pdf import cleanPage


def processInitialPages(script):
    total = []

    for page in script:
        existingY = {}
        for content in page["content"]:
            if content["y"] not in existingY:
                existingY[content["y"]] = True

        total.append(len(existingY))

    avg = sum(total)/len(total)
    firstPages = []
    i = 0
    while i < len(total):
        if total[i] > avg - 10:
            break
        firstPages.append({
            "page": i,
            "content": script[i]["content"]
        })
        i += 1

    firstPages = cleanPage(firstPages, 0)
    firstPages = [x for x in firstPages]
    for page in firstPages:
        page["type"] = "FIRST_PAGES"
    return {
        "firstPages": firstPages,
        "pageStart": i
    }



import json
import argparse

from screenplay_pdf_to_json.parse_pdf import parsePdf, groupDualDialogues, groupSections, sortLines, cleanPage, getTopTrends, stitchSeperateWordsIntoLines, processInitialPages

def convert(scriptFile, pageStart):
    # parse script based on pdfminer.six. Lacking documentation so gotta need some adjustments in our end :(
    newScript = parsePdf(scriptFile)["pdf"]

    firstPagesDict = processInitialPages(newScript)

    firstPages = firstPagesDict["firstPages"]
    skipPage = pageStart if pageStart else firstPagesDict["pageStart"]

    # remove any useless line (page number, empty line, special symbols)
    newScript = cleanPage(newScript, skipPage)

    # sort lines by y. If y is the same, then sort by x
    newScript = sortLines(newScript, skipPage)

    # group dual dialogues into the same segments
    newScript = groupDualDialogues(newScript, skipPage)


    # because of pdfminer's imperfections, we have to stitch words into what's supposed to be part of the same line
    newScript = stitchSeperateWordsIntoLines(newScript, skipPage)

    
    topTrends = getTopTrends(newScript)

    # group into sections based on type
    newScript = groupSections(topTrends, newScript, skipPage)

    newScript = firstPages + newScript
    scriptFile.close()
    return newScript


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description='Parse Screenplay PDF into JSON')

    parser.add_argument('-s', metavar='screenplay', type=str,
                        help='screenplay PDF filename', required=True)

    parser.add_argument('--start', metavar='start page', type=int,
                        help='page to begin analyzing', required=False)

    # start from skipPage set up by user.  default to 0
    args = parser.parse_args()
    scriptFile = open(args.s, 'rb')
    pageStart = args.start
    newScript = convert(scriptFile, pageStart)
    file1 = open('./result.json', 'w+')
    json.dump(newScript, file1, indent=4, ensure_ascii=False)



from copy import copy


def stitchSeperateWordsIntoLines(script, pageStart):
    dialogueStitch = []

    def getJoinedText(textArr):
        return " ".join([x["text"]
                         for x in textArr])
    def segmentTextExists(x): return len(x) > 0 and len(
        x[-1]["text"]) > 0
    for page in script:
        if page["page"] < pageStart:
            continue
        dialogueStitch.append({"page": page["page"], "content": []})

        contentStitch = {
            "segment": []
        }

        for i, content in enumerate(page["content"]):
            if "character2" in content:
                if segmentTextExists(contentStitch["segment"]):
                    dialogueStitch[-1]["content"].append(copy(contentStitch))
                contentStitch = {
                    "segment": []
                }
                dialogueStitch[-1]["content"].append(content)
            elif i > 0 and content["segment"][0]["y"] == page["content"][i-1]["segment"][0]["y"]:
                contentStitch["segment"][-1]["text"] += " " + \
                    getJoinedText(content["segment"])
            else:
                if segmentTextExists(contentStitch["segment"]):
                    dialogueStitch[-1]["content"].append(copy(contentStitch))
                contentStitch = copy(content)

        if len(contentStitch["segment"]) > 0:
            dialogueStitch[-1]["content"].append(copy(contentStitch))

    return dialogueStitch



import operator


def getTopTrends(script):
    trends = {}
    for page in script:
        for section in page["content"]:
            roundedX = round(section["segment"][0]["x"])
            if roundedX not in trends:
                trends[roundedX] = 1
            else:
                trends[roundedX] += 1
    trends = sorted(trends.items(), key=operator.itemgetter(0), reverse=False)
    return trends



import re

headingEnum = ["EXT./INT.", "INT./EXT.", "INT.", "EXT."]


def isHeading(content):
    text = content["text"]
    for heading in headingEnum:
        if not text.endswith(heading) and heading in text:
            return True
    return False


def extractTime(text):
    findTime = re.search(
        '[-,]?[ ]?(DAWN|DUSK|((LATE|EARLY) )?(NIGHT|AFTERNOON|MORNING)|DAYS|DAY|NIGHT|DAYS|DAY|LATER|NIGHT|SAME|CONTINUOUS|(MOMENTS LATER)|SUNSET)|\d{4}', text)

    time = list(filter(lambda x: len(x) > 0, [x.strip(
        "-,. ") for x in text[findTime.start():].split()])) if findTime else None
    return time


def extractHeading(text):
    def stripWord(textArr): return [x.strip() for x in textArr]

    region = re.search(
        '((?:.* )?(?:EXT[\\.]?\\/INT[\\.]?|INT[\\.]?\\/EXT[\\.]?|INT\\.|EXT\\.))', text).groups()[0]
    time = extractTime(text)

    location = text.replace(region, "")
    if time and len(time) > 0:
        location = location[:location.index(time[0])]

    return {
        "region": region,
        "location": location.strip("-,. "),
        "time": time
    }



import re
from screenplay_pdf_to_json.utils.headingHelpers import isHeading


def cleanPage(script, pageStart):
    dialogueStitch = []
    for page in script:
        if page["page"] < pageStart:
            continue
        dialogueStitch.append({"page": page["page"], "content": []})

        firstRound = []
        for content in page["content"]:
            text = content["text"].strip()
            if text == "" or text == "*" or text == "." or text == "\\." or text == "\\" or text == "'":
                continue
            firstRound.append(content)

        for i, content in enumerate(firstRound):
            text = content["text"].strip()
            if not isHeading(content):
                if "TV Calling - For educational purposes only" in text:
                    continue
                elif (re.search(' \d{1,3}[.]?', text) or re.search('\d{1,2}\/\d{1,2}\/\d{2,4}', text)) and (i == 0 or i == len(content)-1):
                    continue
                elif (re.match('(\d|l|i|I){1,3}[.]?(?![\w\d])', text.strip())) and len(text.strip()) < 5:
                    continue
                elif re.search(r"^i{2,3}$", text) and (len(text.split()) < 2 or i == 0 or i == len(content)-1):
                    continue
                elif re.search(r"([(]?CONTINUED[:)]{1,2})", text):
                    continue
                elif re.match('i{2,3}', text.strip()):
                    continue
            dialogueStitch[-1]["content"].append(content)
    removeDuplicates(dialogueStitch)
    return dialogueStitch


def removeDuplicates(script):
    for pageIndex, page in enumerate(script):
        for contentIndex, content in enumerate(page["content"]):
            if contentIndex + 1 < len(page["content"]) and content == page["content"][contentIndex+1]:
                script[pageIndex]["content"].pop(contentIndex+1)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['screenplay_pdf_to_json',
 'screenplay_pdf_to_json.parse_pdf',
 'screenplay_pdf_to_json.utils']

package_data = \
{'': ['*']}

install_requires = \
['pdfminer.six>=20200124,<20200125']

setup_kwargs = {
    'name': 'screenplay-pdf-to-json',
    'version': '0.2.4',
    'description': '',
    'long_description': None,
    'author': 'VVNoodle',
    'author_email': 'brickkace@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)




transitionEnum = [
    "FADE IN",
    "FADE OUT",
    "FADE UP",
    "JUMP CUT",
    "MATCH CUT",
    "SMASH MATCH CUT",
    "MATCH DISSOLVE",
    "CUT",
    "DISSOLVE",
    "FLASH CUT",
    "FREEZE FRAME",
    "IRIS IN",
    "IRIS OUT",
    "WIPE TO"
]


def checkTransition(self, text):
    for transition in transitionEnum:
        if transition in text:
            return True
    return False
